//
//  ServiceCallParams.h
//  CamLoadLib
//
//  Created by Sergey Shramchenko on 09.07.13.
//  Copyright (c) 2013 PhaseOne. All rights reserved.
//

#ifndef CamLoadLib_ServiceCallParams_Client_hpp
#define CamLoadLib_ServiceCallParams_Client_hpp

#include "ServiceCallParams.h"


namespace P1 {
namespace CamLoad {
    namespace ServiceCall {
        template <typename T>
        class Param : public ParamBase
        {
        public:
            // Immutable parameter
            Param(const T& theData)
            :ParamBase(&theData, Loki::TL::IndexOf<KnownTypes, typename PureType<T>::Result>::value)
            {
                // If you see this compilation error
                // please register your parameter type in
                // P1::CamLoad::SerivceCall::KnownTypes typelist.
                STATIC_CHECK((Loki::TL::IndexOf<KnownTypes, typename PureType<T>::Result>::value != -1), Unknown_Parameter_Type_Error);
            }
            // Mutable Parameter
            Param(T* pData)
            :ParamBase(pData, Loki::TL::IndexOf<KnownTypes, typename PureType<T>::Result>::value)
            {
                // If you see this compilation error
                // please register your parameter type in
                // P1::CamLoad::SerivceCall::KnownTypes typelist.
                STATIC_CHECK((Loki::TL::IndexOf<KnownTypes, typename PureType<T>::Result>::value != -1), Unknown_Parameter_Type_Error);
            }
        };
        
        #define BAD_CALL_PARAMETER (ParamBase((const void*)NULL, -1))
        template <class U, class V>
        struct TypeByIndex;
        
        template <class Head, class Tail, /*template<class> class T,*/ typename TParam >
        struct TypeByIndex< ::Loki::Typelist<Head, Tail>, Param<TParam> >
        {
            static ParamBase getParam(ParamBase::TIndex idx, const Param<TParam>& param){
                if(0 == idx)
                    return (SameType<Head, TParam>::value) ? ParamBase(param) : BAD_CALL_PARAMETER;
                else
                    return TypeByIndex<Tail, Param<TParam> >::getParam(idx - 1, param);
            }
            
        };
        
        template <template<class> class T, class TParam >
        struct TypeByIndex< ::Loki::NullType, T<TParam> >
        {
            static ParamBase getParam(ParamBase::TIndex, const T<TParam>&)
            {
                return BAD_CALL_PARAMETER;
            }
        };
        template <template<class> class T, class TParam>
        void TypeSafeParams::addParam(const T<TParam>& param) {
            if(m_TypeInformer) {
                // Check provided parameter against registered callback
                ParamBase::TIndex newParamIndex=  m_Params.size();
                short paramTypeIndex = m_TypeInformer->paramTypeAtIndex(newParamIndex);
                ParamBase baseParam = TypeByIndex<KnownTypes, T<TParam> >::getParam(paramTypeIndex, param);
                m_Params.push_back(baseParam);
            } else {
                // If we can't check callback parameter type
                // just add the parameter without checking
                add(param);
            }
        }

        class NativeParameterBase {
        public:
            virtual ~NativeParameterBase() {/*NSAssert((m_Buffer == NULL), @"NativeParameterBase memory leak!");*/}
        protected:
            NativeParameterBase()
            :m_Buffer(NULL)
            {}
            void* m_Buffer;
        private:
            NativeParameterBase(const NativeParameterBase&);
            NativeParameterBase& operator=(const NativeParameterBase&);
        };
        
        template <typename T>
        class NativeParameter : public NativeParameterBase {
        public:
            typedef typename P1::CamLoad::ServiceCall::PureType<T>::Result TParam;
            
            NativeParameter() {
                m_Buffer = new TParam();
            }
            NativeParameter(const TParam& p) {
                m_Buffer = new TParam(p);
            }
            ~NativeParameter() {
                delete (T*)m_Buffer;
                m_Buffer = NULL;
            }
            operator const T&() const {
                return *((T*)m_Buffer);
            }
            operator T*() {
                return (T*)m_Buffer;
            }
            
        };
        

        
        
    }// ServiceCall namespace
}} // namescpace

#endif
